/*
 * Copyright 2011, 2016, 2017 Sony Corporation
 */
#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/cdev.h>
#include <linux/uaccess.h>
#include <linux/poll.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/list.h>
#include <linux/usb/gadget.h>
#include <linux/icx_usbconn.h>
#include <linux/icx_usbtrns.h>

#define MODULE_NAME "icx_usbconn"

#define USE_INFOMATION_MESSAGE
/* #define USE_DEBUG_MESSAGE */

#define ERR_MSG(format, arg...) pr_err("(E)%s:%s(%d) " format "\n", \
					MODULE_NAME, __func__, __LINE__, ## arg)
#define WRN_MSG(format, arg...) pr_warn("(W)%s:%s(%d) " format "\n", \
					MODULE_NAME, __func__, __LINE__, ## arg)

#ifdef USE_INFOMATION_MESSAGE
#define INF_MSG(format, arg...) pr_info("(I)%s:%s(%d) " format "\n", \
					MODULE_NAME, __func__, __LINE__, ## arg)
#else
#define INF_MSG(format, arg...) do {} while (0)
#endif

#ifdef USE_DEBUG_MESSAGE
#define DBG_MSG(format, arg...) pr_warn("(D)%s:%s(%d) " format "\n", \
					MODULE_NAME, __func__, __LINE__, ## arg)
#else
#define DBG_MSG(format, arg...) do {} while (0)
#endif


/* USBCONN */
#define USB_CONN_DRV_MAJOR      0
#define USB_CONN_DRV_MINOR      0
#define USB_CONN_CDEV_FILE_NAME "usbconn"

struct usbconn_dev {
	int                 users;
	struct semaphore    sem;
	struct cdev         cdev;
	wait_queue_head_t   wqh;
	spinlock_t          lock;
	struct msc_to_conn_func *msc_drv;
	int                 connect_sts;
	int                 acc_sts;
	struct list_head    list;
};

struct usbconn_connect_status {
	struct list_head    entry;
	int                 connect_sts;
};

static int  usbconn_init(void);
static void usbconn_exit(void);
static int  usbconn_open(struct inode *inode, struct file *filp);
static int  usbconn_release(struct inode *inode, struct file *filp);
static long usbconn_ioctl(struct file *filp,
			  unsigned int cmd, unsigned long arg);
static unsigned int usbconn_poll(struct file *file, poll_table *wait);

static int usbconn_major = USB_CONN_DRV_MAJOR;
static int usbconn_minor = USB_CONN_DRV_MINOR;

static struct usbconn_dev *usbconn_devices;
const struct file_operations usbconn_fops = {
	.owner   = THIS_MODULE,
	.unlocked_ioctl   = usbconn_ioctl,
	.poll    = usbconn_poll,
	.open    = usbconn_open,
	.release = usbconn_release,
};

/* called from gadget */
static struct conn_to_gadget_func conn_to_gadget_func_tbl = {
	.change_status = usbconn_chg_sts,
};

/* called from msc */
static struct conn_to_msc_func conn_to_msc_func_tbl = {
	.change_status = usbconn_chg_sts,
};

/* registered from msc */
static struct msc_to_conn_func msc_to_conn_func_tbl = {
	.allow_acc = NULL,
};

struct class *usbconn_class;

module_param(usbconn_major, int, S_IRUGO);
module_param(usbconn_minor, int, S_IRUGO);


/* USBTRNS */
#define USB_TRNS_DRV_MAJOR      0
#define USB_TRNS_DRV_MINOR      0
#define USB_TRNS_CDEV_FILE_NAME "usbtrns"

struct usbtrns_dev {
	int                 users;
	struct semaphore    sem;
	struct cdev         cdev;
	wait_queue_head_t   wqh;
	spinlock_t          lock;
	struct msc_to_trns_func *msc_drv;
	int                 wake_up_flg;
};

static int  usbtrns_open(struct inode *inode, struct file *filp);
static int  usbtrns_release(struct inode *inode, struct file *filp);
static int  usbtrns_ioctl(struct file *filp,
			  unsigned int cmd, unsigned long arg);
static unsigned int usbtrns_poll(struct file *file, poll_table *wait);
static void usbtrns_start(void);

int usbtrns_major = USB_TRNS_DRV_MAJOR;
int usbtrns_minor = USB_TRNS_DRV_MINOR;

struct usbtrns_dev *usbtrns_devices;
const struct file_operations usbtrns_fops = {
	.owner   =  THIS_MODULE,
	.unlocked_ioctl     = usbtrns_ioctl,
	.poll    =  usbtrns_poll,
	.open    =  usbtrns_open,
	.release =  usbtrns_release,
};

/* called from msc */
static struct trns_to_msc_func trns_to_msc_func_tbl = {
	.trns_start = usbtrns_start,
};

/* registered from msc */
static struct msc_to_trns_func msc_to_trns_func_tbl = {
	.get_lapsed_time = NULL,
	.fsync_all = NULL,
};

module_param(usbtrns_major, int, S_IRUGO);
module_param(usbtrns_minor, int, S_IRUGO);

MODULE_LICENSE("GPL");


/*== Init Entry            ==*/
static int usbconn_init(void)
{
	struct device *usbconn_device = NULL;
	struct device *usbtrns_device = NULL;
	dev_t  dev = 0;
	int    ret = 0;
	bool   status = false;

	DBG_MSG("usbconn initialize start");

	/*== USBCONN INITIALIZE PROCESS ==*/
	if (usbconn_major != 0) {
		dev = MKDEV(usbconn_major, usbconn_minor);
		ret = register_chrdev_region(dev, 1, USB_CONN_CDEV_FILE_NAME);
		DBG_MSG("major(%d) minor(%d)", usbconn_major, usbconn_minor);
	} else {
		ret = alloc_chrdev_region(&dev, usbconn_minor, 1,
					  USB_CONN_CDEV_FILE_NAME);
		usbconn_major = MAJOR(dev);
		DBG_MSG("major(%d) minor(%d)", usbconn_major, usbconn_minor);
	}

	if (ret < 0) {
		ERR_MSG("can't get major(%d)", usbconn_major);
		goto error_ret;
	}

	usbconn_devices = kmalloc(sizeof(struct usbconn_dev), GFP_KERNEL);
	if (usbconn_devices == NULL) {
		ret = -ENOMEM;
		goto error_ret;
	}

	memset(usbconn_devices, 0, sizeof(struct usbconn_dev));

	usbconn_devices->users = 0;
	sema_init(&usbconn_devices->sem, 1);
	spin_lock_init(&usbconn_devices->lock);
	init_waitqueue_head(&usbconn_devices->wqh);

	usbconn_devices->acc_sts = USB_CONN_ACC_INIT;

	cdev_init(&usbconn_devices->cdev, &usbconn_fops);
	usbconn_devices->cdev.owner = THIS_MODULE;
	usbconn_devices->cdev.ops = &usbconn_fops;

	INIT_LIST_HEAD(&usbconn_devices->list);

	ret = cdev_add(&usbconn_devices->cdev, dev, 1);
	if (ret)
		ERR_MSG("failed adding usbconn(%d)", ret);

	usbconn_class = class_create(THIS_MODULE, "icx_usbconn");
	if (IS_ERR(usbconn_class)) {
		ERR_MSG("error class_create");
		ret = -ENOMEM;
		goto error_ret;
	}

	usbconn_device = device_create(usbconn_class, NULL,
				       dev, NULL, "usbconn");

	usbconn_devices->msc_drv = &msc_to_conn_func_tbl;
#ifndef CONFIG_USBCONN_CHARGING_MODE_ONLY
	msc_bind_usbconn(&msc_to_conn_func_tbl, &conn_to_msc_func_tbl);
#endif

#ifdef CONFIG_ENABLE_USBCONN
	gadget_bind_usbconn(&conn_to_gadget_func_tbl);

	status = get_usb_connect_status();
#endif
	if (status == true) {
		usbconn_devices->connect_sts = USB_CONNECT;
		ERR_MSG("initial connect status(connect)");
	} else {
		usbconn_devices->connect_sts = USB_DISCONNECT;
		ERR_MSG("initial connect status(disconnect)");
	}

	/*== USBTRNS INITIALIZE PROCESS ==*/
	dev = 0;

	DBG_MSG("usbtrns initialize start");

	if (usbtrns_major != 0) {
		dev = MKDEV(usbtrns_major, usbtrns_minor);
		ret = register_chrdev_region(dev, 1, USB_TRNS_CDEV_FILE_NAME);
		DBG_MSG("major(%d) minor(%d)", usbtrns_major, usbtrns_minor);
	} else {
		ret = alloc_chrdev_region(&dev, usbtrns_minor, 1,
					  USB_TRNS_CDEV_FILE_NAME);
		usbtrns_major = MAJOR(dev);
		DBG_MSG("usbtrns2: major=%d minor=%d",
			 usbtrns_major, usbtrns_minor);
	}

	if (ret < 0) {
		ERR_MSG("can't get major(%d)", usbtrns_major);
		goto error_ret;
	}

	usbtrns_devices = kmalloc(sizeof(struct usbtrns_dev), GFP_KERNEL);
	if (usbtrns_devices == NULL) {
		ret = -ENOMEM;
		goto error_ret;
	}

	memset(usbtrns_devices, 0, sizeof(struct usbtrns_dev));

	usbtrns_devices->users = 0;
	sema_init(&usbtrns_devices->sem, 1);
	spin_lock_init(&usbtrns_devices->lock);
	init_waitqueue_head(&usbtrns_devices->wqh);

	usbtrns_devices->wake_up_flg = 0;

	cdev_init(&usbtrns_devices->cdev, &usbtrns_fops);
	usbtrns_devices->cdev.owner = THIS_MODULE;
	usbtrns_devices->cdev.ops   = &usbtrns_fops;

	ret = cdev_add(&usbtrns_devices->cdev, dev, 1);
	if (ret)
		ERR_MSG("failed adding usbtrns(%d)", ret);

	usbtrns_device = device_create(usbconn_class,
				       NULL, dev, NULL, "usbtrns");

#ifndef CONFIG_USBCONN_CHARGING_MODE_ONLY
	msc_bind_usbtrns(&msc_to_trns_func_tbl, &trns_to_msc_func_tbl);
#endif
	usbtrns_devices->msc_drv = &msc_to_trns_func_tbl;

	DBG_MSG("exit");
	return 0;

error_ret:
	usbconn_exit();
	return ret;
}

/*== Exit Entry            ==*/
static void usbconn_exit(void)
{
	dev_t devno;

	DBG_MSG("start");

	/*== USBCONN EXIT PROCESS ==*/
	devno = MKDEV(usbconn_major, usbconn_minor);

	if (usbconn_devices) {
		struct usbconn_connect_status *status, *status_next;
		list_for_each_entry_safe(status, status_next,
					 &usbconn_devices->list, entry) {
			list_del(&status->entry);
			kfree(status);
		}
		device_destroy(usbconn_class, devno);
		cdev_del(&usbconn_devices->cdev);
		kfree(usbconn_devices);
	}

	unregister_chrdev_region(devno, 1);

	/*== USBTRNS EXIT PROCESS ==*/
	devno = MKDEV(usbtrns_major, usbtrns_minor);

	if (usbtrns_devices) {
		device_destroy(usbconn_class, devno);
		cdev_del(&usbtrns_devices->cdev);
		kfree(usbtrns_devices);
	}

	unregister_chrdev_region(devno, 1);

	class_destroy(usbconn_class);

	DBG_MSG("exit");
}

/*== Operation Function    ==*/
static int usbconn_open(struct inode *inode, struct file *filp)
{
	struct usbconn_dev *dev;

	DBG_MSG("start(inode:%p filp:%p)", inode, filp);

	dev = container_of(inode->i_cdev, struct usbconn_dev, cdev);
	filp->private_data = dev;

	if (down_interruptible(&dev->sem))
		return -ERESTARTSYS;

	dev->users++;
	DBG_MSG("users(%d)", dev->users);

	up(&dev->sem);

	DBG_MSG("exit");
	return 0;
}

static int usbconn_release(struct inode *inode, struct file *filp)
{
	struct usbconn_dev *dev;

	DBG_MSG("start(inode:%p filp:%p)", inode, filp);

	dev = filp->private_data;

	if (down_interruptible(&dev->sem))
		return -ERESTARTSYS;

	dev->users--;
	DBG_MSG("users(%d)", dev->users);

	up(&dev->sem);

	DBG_MSG("exit");
	return 0;
}

static long usbconn_ioctl(struct file *filp,
			  unsigned int cmd, unsigned long arg)
{
	struct usbconn_dev *dev = (struct usbconn_dev *)filp->private_data;
	unsigned long flags;
	long ret = 0;
	unsigned int  sts = USB_DISCONNECT;

	DBG_MSG("start(filp:%p cmd:%d now:%d)", filp, cmd, dev->connect_sts);

	if (_IOC_TYPE(cmd) != MSC_CONN_IOC_MAGIC)
		return -ENOTTY;
	if (_IOC_NR(cmd) > MSC_CONN_IOC_MAXNR)
		return -ENOTTY;

	switch (cmd) {
	case MSC_GET_CSTAT:
		DBG_MSG("recv GET_CSTAT");

		spin_lock_irqsave(&dev->lock, flags);

		if (!list_empty(&dev->list)) {
			struct usbconn_connect_status *status =
				list_first_entry(&dev->list,
					struct usbconn_connect_status, entry);
			if (status) {
				sts = status->connect_sts;
				list_del(&status->entry);
				kfree(status);

				if (!list_empty(&dev->list))
					WRN_MSG("list has more entry");
			} else {
				sts = dev->connect_sts;
			}
		} else {
			sts = dev->connect_sts;
		}

		if (sts != USB_DISCONNECT &&
		    dev->acc_sts == USB_CONN_ACC_PROHIBIT) {
			DBG_MSG("change the status");
			sts = USB_DISCONNECT;
		}

		spin_unlock_irqrestore(&dev->lock, flags);

		INF_MSG("status(notice:%s now:%s)",
			sts == USB_CONNECT    ? "CONNECT" :
			sts == USB_DISCONNECT ? "DISCONNECT" : "FWUPDATE",
			dev->connect_sts == USB_CONNECT    ? "CONNECT" :
			dev->connect_sts == USB_DISCONNECT ?
						"DISCONNECT" : "FWUPDATE");

		ret = copy_to_user((void *)arg, &sts, sizeof(sts));
		break;

	case MSC_ALLOW_ACC:
		DBG_MSG("recv ALLOW_ACC (now:%d)", dev->acc_sts);

		dev->acc_sts = USB_CONN_ACC_ALLOW;
		if (dev->msc_drv->allow_acc != NULL) {
			dev->msc_drv->allow_acc(USB_CONN_ACC_ALLOW);
			DBG_MSG("notice allow status(ALLOW)");
		}

		DBG_MSG("set ALLOW_ACC (now:%d)", dev->acc_sts);
		break;

	case MSC_PROHIBIT_ACC:
		DBG_MSG("recv PROHIBIT_ACC (now:%d)", dev->acc_sts);

		dev->acc_sts = USB_CONN_ACC_PROHIBIT;
		if (dev->msc_drv->allow_acc != NULL) {
			dev->msc_drv->allow_acc(USB_CONN_ACC_PROHIBIT);
			DBG_MSG("notice allow status(PROHIBIT)");
		}

		DBG_MSG("set PROHIBIT_ACC (now:%d)", dev->acc_sts);
		break;

	default:
		ERR_MSG("invalid command(%d)", cmd);
		ret = -ENOTTY;
		break;
	}

	DBG_MSG("exit(now:%d)", dev->connect_sts);
	return ret;
}

static unsigned int usbconn_poll(struct file *file, poll_table *wait)
{
	struct usbconn_dev *dev = (struct usbconn_dev *)file->private_data;
	unsigned long flags;
	int ret = 0;

	DBG_MSG("start(file:%p wait:%p)", file, wait);

	poll_wait(file, &dev->wqh, wait);

	spin_lock_irqsave(&dev->lock, flags);

	if (!list_empty(&dev->list)) {
		ret |= POLLIN | POLLRDNORM;
		DBG_MSG("exist data(%d)", ret);
	}

	spin_unlock_irqrestore(&dev->lock, flags);

	DBG_MSG("exit(%d)", ret);
	return ret;
}

/*== USBCONN API (set the state)   ==*/
int usbconn_chg_sts(int sts)
{
	struct usbconn_dev *dev = usbconn_devices;

	DBG_MSG("start(now:%d)", dev->connect_sts);

	if (sts != dev->connect_sts) {
		INF_MSG("update status(set:%s now:%s)",
			sts == USB_CONNECT    ? "CONNECT" :
			sts == USB_DISCONNECT ? "DISCONNECT" : "FWUPDATE",
			dev->connect_sts == USB_CONNECT    ? "CONNECT" :
			dev->connect_sts == USB_DISCONNECT ?
						"DISCONNECT" : "FWUPDATE");

		/* update USB connect status */
		dev->connect_sts = sts;
		{
			struct usbconn_connect_status *status =
				kmalloc(sizeof(struct usbconn_connect_status),
					GFP_KERNEL);
			if (status) {
				status->connect_sts = sts;
				list_add_tail(&status->entry, &dev->list);

				DBG_MSG("status(%d) queuing", sts);
			} else {
				ERR_MSG("no memory");
				return -ENOMEM;
			}
		}

		if (dev->connect_sts == USB_CONNECT) {
			DBG_MSG("usb connect");

			dev->acc_sts = USB_CONN_ACC_INIT;
			if (dev->msc_drv->allow_acc != NULL) {
				dev->msc_drv->allow_acc(USB_CONN_ACC_INIT);
				DBG_MSG("notice allow status(ACC_INIT)");
			}
		}

		DBG_MSG("wakeup work queue");
		wake_up_interruptible(&dev->wqh);
	} else {
		DBG_MSG("no change(%d)", sts);
	}

	DBG_MSG("exit");
	return 0;
}
EXPORT_SYMBOL(usbconn_chg_sts);


/*== USBTRNS Operation Function    ==*/
static int usbtrns_open(struct inode *inode, struct file *filp)
{
	struct usbtrns_dev *dev;

	DBG_MSG("start(inode:%p filp:%p)", inode, filp);

	dev = container_of(inode->i_cdev, struct usbtrns_dev, cdev);
	filp->private_data = dev;

	if (down_interruptible(&dev->sem))
		return -ERESTARTSYS;

	dev->users++;
	DBG_MSG("users(%d)", dev->users);

	up(&dev->sem);

	DBG_MSG("exit");
	return 0;
}

static int usbtrns_release(struct inode *inode, struct file *filp)
{
	struct usbtrns_dev *dev;

	DBG_MSG("start(inode:%p filp:%p)", inode, filp);

	dev = filp->private_data;

	if (down_interruptible(&dev->sem))
		return -ERESTARTSYS;

	dev->users--;
	DBG_MSG("users(%d)", dev->users);

	up(&dev->sem);

	DBG_MSG("exit");
	return 0;
}

static int usbtrns_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
	struct usbtrns_dev *dev = (struct usbtrns_dev *)filp->private_data;
	struct usbconn_dev *usbconn_dev_info = usbconn_devices;
	unsigned long lapsed_time = 0xFFFFFFFF;
	int ret = 0, tmp = 0;

	DBG_MSG("start(filp:%p cmd:%d now:%d)",
		 filp, cmd, usbconn_dev_info->connect_sts);

	if (_IOC_TYPE(cmd) != MSC_TRNS_IOC_MAGIC)
		return -ENOTTY;
	if (_IOC_NR(cmd) > MSC_TRNS_IOC_MAXNR)
		return -ENOTTY;

	switch (cmd) {
	case MSC_GET_LAPSED_TIME:
		DBG_MSG("recv GET_LAPSED_TIME");

		if (usbconn_dev_info->connect_sts != USB_DISCONNECT) {
			if (dev->msc_drv->get_lapsed_time != NULL) {
				ret = dev->msc_drv->get_lapsed_time(
								&lapsed_time);
				INF_MSG("get the lapsed_time(%ld)",
					 lapsed_time);
			} else {
				ERR_MSG("not regist");
				ret = -1;
			}
		} else {
			ERR_MSG("status is disconnect");
			ret = -1;
		}

		if (ret != 0) {
			ERR_MSG("can not get the lapsed_time...");
			lapsed_time = 0xFFFFFFFF;
		}

		tmp = copy_to_user((void *)arg, &lapsed_time,
				    sizeof(lapsed_time));
		if (tmp != 0)
			ret = tmp;
		break;

	case MSC_FSYNC:
		DBG_MSG("recv MSC_FSYNC");

		if (usbconn_dev_info->connect_sts != USB_DISCONNECT) {
			if (dev->msc_drv->fsync_all != NULL) {
				dev->msc_drv->fsync_all();
				INF_MSG("call regist function");
			} else {
				ERR_MSG("not regist");
				ret = -1;
			}
		} else {
			ERR_MSG("status is disconnect");
			ret = -1;
		}
		break;

	default:
		ERR_MSG("invalid command(%d)", cmd);
		ret = -ENOTTY;
		break;
	}

	DBG_MSG("exit(ret:%d)", ret);
	return ret;
}

static unsigned int usbtrns_poll(struct file *file, poll_table *wait)
{
	struct usbtrns_dev *dev = (struct usbtrns_dev *)file->private_data;
	unsigned long flags;
	int ret = 0;

	DBG_MSG("start(file:%p wait:%p)", file, wait);

	poll_wait(file, &dev->wqh, wait);

	spin_lock_irqsave(&dev->lock, flags);

	if (wait != 0) {
		dev->wake_up_flg = 0;
		DBG_MSG("call from application");
	} else {
		if (dev->wake_up_flg != 0) {
			ret |= POLLIN | POLLRDNORM;
			DBG_MSG("call from driver");
		}
	}

	spin_unlock_irqrestore(&dev->lock, flags);

	DBG_MSG("exit(%d)", ret);
	return ret;
}


/*== USBTRNS API (start transfer)  ==*/
static void usbtrns_start(void)
{
	struct usbtrns_dev *dev = usbtrns_devices;
	unsigned long flags;

	DBG_MSG("start(now:%d)", dev->wake_up_flg);

	spin_lock_irqsave(&dev->lock, flags);

	if (dev->wake_up_flg == 0) {
		dev->wake_up_flg = 1;
		DBG_MSG("usbtrns wake_up");
		wake_up_interruptible(&dev->wqh);
	}

	spin_unlock_irqrestore(&dev->lock, flags);
}


module_init(usbconn_init);
module_exit(usbconn_exit);

